from typing import Any, Dict, ItemsView, KeysView, ValuesView
from urllib.parse import unquote_plus


def build_dict_from_path(path: str, value) -> Dict[str, Any]:
    """
    Creates dictionary representing passed path with given value.
    For example: [some][path] will be turned to {"some":{"path": value}} dict.
    :param path:
    :param value:
    :return:
    """
    starting_bracket = path.find("[")
    if starting_bracket == 0:
        raise ValueError("Path cannot start with [")
    if path[-1:] != "]":
        return {path: value}
    parsed_path = [path[:starting_bracket]]
    parsed_path = parsed_path + path[starting_bracket + 1 : -1].split("][")

    for part in parsed_path:
        if "[" in part or "]" in part:
            return {path: value}

    def _create_leaf(_parsed_path: list):
        if len(_parsed_path) == 1:
            if not _parsed_path[0]:
                return [value]
            else:
                return {_parsed_path[0]: value}
        if not _parsed_path[0]:
            return [_create_leaf(_parsed_path[1:])]
        else:
            return {_parsed_path[0]: _create_leaf(_parsed_path[1:])}

    return _create_leaf(parsed_path)


def deep_merge(a: Dict[str, Any], b: Dict[str, Any]) -> Dict[str, Any]:
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                a[key] = deep_merge(a[key], b[key])
            elif isinstance(a[key], list) and isinstance(b[key], list):
                a[key] = a[key] + b[key]
            elif isinstance(b[key], list):
                a[key] = [a[key]] + b[key]
            elif isinstance(b[key], dict):
                a[key] = {"": a[key], **b[key]}
            else:
                a[key] = [a[key], b[key]]
        else:
            a[key] = b[key]
    return a


def parse_qs(query: str) -> Dict[str, Any]:
    """
    Parse query string with json forms support, more available in the following link
    https://www.w3.org/TR/html-json-forms/
    :param query:
    :return:
    """
    result: Dict[str, Any] = {}
    if query == "":
        return result

    for item in query.split("&"):
        (name, value) = item.split("=")
        value = parse_qs_value(value)
        name = unquote_plus(name)
        if "[" in name:
            result = deep_merge(result, build_dict_from_path(name, value))
        elif name in result:
            if isinstance(result[name], list):
                result[name].append(value)
            else:
                result[name] = [result[name], value]
        else:
            result[name] = value

    return result


def parse_qs_value(value: str) -> Any:
    value = unquote_plus(value)
    if value == "true":
        return True

    if value == "false":
        return False

    try:
        return int(value)
    except ValueError:
        pass

    try:
        return float(value)
    except ValueError:
        pass

    return value


class HttpQueryString:
    def __init__(self, string: str):
        self._str = string
        self._params = parse_qs(string)

    def __getitem__(self, key) -> str:
        return self.get(key)

    def __contains__(self, key) -> bool:
        return key in self._params

    def __repr__(self):
        return self._str

    def __str__(self) -> str:
        return self._str

    def get(self, key: str, default: Any = None):
        return self._params.get(key, default)

    def items(self) -> ItemsView:
        return self._params.items()

    def values(self) -> ValuesView:
        return self._params.values()

    def keys(self) -> KeysView:
        return self._params.keys()

    def __eq__(self, other) -> bool:
        if not isinstance(other, HttpQueryString):
            return False

        return other._str == self._str

    def __iter__(self):
        return iter(self._params)


__all__ = ["HttpQueryString", "build_dict_from_path", "parse_qs", "parse_qs_value"]



from cheroot.wsgi import Server
from typing import Callable


def wsgi_serve(
    wsgi_handler: Callable,
    host: str = "127.0.0.1",
    port: int = 80,
    workers: int = 1,
) -> None:
    wsgi_server = Server((host, port), wsgi_handler, numthreads=workers)
    wsgi_server.start()



from enum import Enum
from io import BytesIO
from typing import Any, Callable, Dict

from chocs.application import Application
from chocs.http.http_error import HttpError
from chocs.http.http_headers import HttpHeaders
from chocs.http.http_method import HttpMethod
from chocs.http.http_query_string import HttpQueryString
from chocs.http.http_request import HttpRequest
from chocs.http.http_response import HttpResponse


def create_http_request_from_wsgi(environ: Dict[str, Any]) -> HttpRequest:
    headers = HttpHeaders()
    for key, value in environ.items():
        if not key.startswith("HTTP"):
            continue
        headers.set(key, value)
    headers.set("Content-Type", environ.get("CONTENT_TYPE", "text/plain"))

    if "wsgi.input" in environ:  # unify all the different types of wsgi server implementations
        body = BytesIO(environ["wsgi.input"].read())
    else:
        body = BytesIO(b"")

    return HttpRequest(
        method=HttpMethod(environ.get("REQUEST_METHOD", "GET").upper()),
        path=environ.get("PATH_INFO", "/"),
        body=body,
        query_string=HttpQueryString(environ.get("QUERY_STRING", "")),
        headers=headers,
    )


def create_wsgi_handler(
    application: Application, debug: bool = False
) -> Callable[[Dict[str, Any], Callable[..., Any]], BytesIO]:
    def _handler(environ: Dict[str, Any], start: Callable) -> BytesIO:
        request = create_http_request_from_wsgi(environ)
        if debug:
            try:
                response = application(request)
            except HttpError as http_error:
                response = HttpResponse(http_error.http_message, http_error.status_code)
        else:
            # Always send a response
            try:
                response = application(request)
            except HttpError as http_error:
                response = HttpResponse(http_error.http_message, http_error.status_code)
            except Exception:
                response = HttpResponse("Internal Server Error", 500)

        headers = response.headers
        for cookie in response.cookies.values():
            headers.set("Set-Cookie", cookie.serialise())

        start(
            str(int(response.status_code)),
            [(key, value) for key, value in headers.items()],
        )

        response.body.seek(0)
        return response.body

    return _handler


class WsgiServers(Enum):
    DEFAULT = "bjoern"
    GUNICORN = "gunicorn"
    BJOERN = "bjoern"
    CHERRYPY = "cherrypy"


def serve(
    application: Application,
    host: str = "127.0.0.1",
    port: int = 80,
    workers: int = 1,
    debug: bool = False,
    wsgi_server: WsgiServers = WsgiServers.DEFAULT,
) -> None:

    wsgi_handler = create_wsgi_handler(application, debug=debug)
    wsgi_options = {
        "host": host,
        "port": port,
        "workers": workers,
    }

    if wsgi_server == WsgiServers.BJOERN:
        try:
            from .bjoern_support import wsgi_serve  # type: ignore
        except ImportError:
            raise RuntimeError("`bjoern` package must be installed before using `chocs.serve`.")

    elif wsgi_server == WsgiServers.GUNICORN:
        try:
            from .gunicorn_support import wsgi_serve  # type: ignore
        except ImportError:
            raise RuntimeError("`gunicorn` package must be installed before using `chocs.serve`.")

    elif wsgi_server == WsgiServers.CHERRYPY:
        try:
            from .cherrypy_support import wsgi_serve  # type: ignore
        except ImportError:
            raise RuntimeError("`cheroot` package must be installed before using `chocs.serve`.")

    else:
        raise RuntimeError("Unsupported wsgi server")

    wsgi_serve(wsgi_handler, **wsgi_options)  # type: ignore



from typing import Callable

from chocs.http.http_request import HttpRequest
from chocs.http.http_response import HttpResponse

HttpHandlerFunction = Callable[[HttpRequest], HttpResponse]


__all__ = ["HttpHandlerFunction"]



from chocs.http import *
from .application import Application
from .middleware.application_middleware import ApplicationMiddleware
from .routing import Route, Router
from .wsgi.wsgi_support import WsgiServers, create_wsgi_handler, serve



from enum import Enum
from typing import Union


class HttpStatus(Enum):
    CONTINUE = 100, "Continue"
    SWITCHING_PROTOCOLS = 101, "Switching Protocols"
    PROCESSING = 102, "Processing"
    OK = 200, "OK"
    CREATED = 201, "Created"
    ACCEPTED = 202, "Accepted"
    NON_AUTHORITATIVE_INFORMATION = 203, "Non-Authoritative Information"
    NO_CONTENT = 204, "No Content"
    RESET_CONTENT = 205, "Reset Content"
    PARTIAL_CONTENT = 206, "Partial Content"
    MULTI_STATUS = 207, "Multi-Status"
    ALREADY_REPORTED = 208, "Already Reported"
    IM_USED = 226, "IM Used"
    MULTIPLE_CHOICES = 300, "Multiple Choices"
    MOVED_PERMANENTLY = 301, "Moved Permanently"
    FOUND = 302, "Found"
    SEE_OTHER = 303, "See Other"
    NOT_MODIFIED = 304, "Not Modified"
    USE_PROXY = 305, "Use Proxy"
    TEMPORARY_REDIRECT = 307, "Temporary Redirect"
    PERMANENT_REDIRECT = 308, "Permanent Redirect"
    BAD_REQUEST = 400, "Bad Request"
    UNAUTHORIZED = 401, "Unauthorized"
    PAYMENT_REQUIRED = 402, "Payment Required"
    FORBIDDEN = 403, "Forbidden"
    NOT_FOUND = 404, "Not Found"
    METHOD_NOT_ALLOWED = 405, "Method Not Allowed"
    NOT_ACCEPTABLE = 406, "Not Acceptable"
    PROXY_AUTHENTICATION_REQUIRED = 407, "Proxy Authentication Required"
    REQUEST_TIMEOUT = 408, "Request Timeout"
    CONFLICT = 409, "Conflict"
    GONE = 410, "Gone"
    LENGTH_REQUIRED = 411, "Length Required"
    PRECONDITION_FAILED = 412, "Precondition Failed"
    REQUEST_ENTITY_TOO_LARGE = 413, "Request Entity Too Large"
    REQUEST_URI_TOO_LONG = 414, "Request-URI Too Long"
    UNSUPPORTED_MEDIA_TYPE = 415, "Unsupported Media Type"
    REQUESTED_RANGE_NOT_SATISFIABLE = 416, "Requested Range Not Satisfiable"
    EXPECTATION_FAILED = 417, "Expectation Failed"
    MISDIRECTED_REQUEST = 421, "Misdirected Request"
    UNPROCESSABLE_ENTITY = 422, "Unprocessable Entity"
    LOCKED = 423, "Locked"
    FAILED_DEPENDENCY = 424, "Failed Dependency"
    UPGRADE_REQUIRED = 426, "Upgrade Required"
    PRECONDITION_REQUIRED = 428, "Precondition Required"
    TOO_MANY_REQUESTS = 429, "Too Many Requests"
    REQUEST_HEADER_FIELDS_TOO_LARGE = 431, "Request Header Fields Too Large"
    UNAVAILABLE_FOR_LEGAL_REASONS = 451, "Unavailable For Legal Reasons"
    INTERNAL_SERVER_ERROR = 500, "Internal Server Error"
    NOT_IMPLEMENTED = 501, "Not Implemented"
    BAD_GATEWAY = 502, "Bad Gateway"
    SERVICE_UNAVAILABLE = 503, "Service Unavailable"
    GATEWAY_TIMEOUT = 504, "Gateway Timeout"
    HTTP_VERSION_NOT_SUPPORTED = 505, "HTTP Version Not Supported"
    VARIANT_ALSO_NEGOTIATES = 506, "Variant Also Negotiates"
    INSUFFICIENT_STORAGE = 507, "Insufficient Storage"
    LOOP_DETECTED = 508, "Loop Detected"
    NOT_EXTENDED = 510, "Not Extended"
    NETWORK_AUTHENTICATION_REQUIRED = 511, "Network Authentication Required"

    def __str__(self) -> str:
        return "%d %s" % self.value

    def __int__(self) -> int:
        return self.value[0]

    def __lt__(self, other: Union[int, "HttpStatus"]) -> bool:
        if isinstance(other, int):
            return self.value[0] < other

        return self.value[0] < other.value[0]

    def __le__(self, other: Union[int, "HttpStatus"]) -> bool:
        if isinstance(other, int):
            return self.value[0] <= other

        return self.value[0] <= other.value[0]

    def __gt__(self, other: Union[int, "HttpStatus"]) -> bool:
        if isinstance(other, int):
            return self.value[0] > other

        return self.value[0] > other.value[0]

    def __ge__(self, other: Union[int, "HttpStatus"]) -> bool:
        if isinstance(other, int):
            return self.value[0] >= other

        return self.value[0] >= other.value[0]

    def __eq__(self, other: Union[int, "HttpStatus"]) -> bool:  # type: ignore
        if isinstance(other, int):
            return self.value[0] == other

        return self.value[0] == other.value[0]

    @classmethod
    def from_int(cls, status: int) -> "HttpStatus":
        value = [item for item in cls if item.value[0] == status]
        if value:
            return value[0]

        raise ValueError(f"Invalid status value `{status}`")


__all__ = ["HttpStatus"]



from cgi import parse_header
from enum import Enum
from tempfile import TemporaryFile
from typing import IO, Any, Dict, Tuple


class UploadedFile:
    """
    Proxy class for TemporaryFile (uploaded file)
    """

    def __init__(self, file: IO[Any], mimetype: str, filename: str):
        self.file = file
        self.mimetype = mimetype
        self.filename = filename
        self.length = 0
        self._str = ""

    def read(self) -> bytes:
        return self.file.read()

    def seek(self, offset: int) -> int:
        return self.file.seek(offset)

    def close(self) -> None:
        self.file.close()

    def save(self, path: str) -> IO[Any]:
        if self.file.closed:
            raise ValueError(f"Cannot save to file {path} of closed stream.")
        with open(path, "wb") as file:
            self.seek(0)
            file.write(self.read())

        return file

    def __float__(self) -> None:
        raise ValueError(f"Cannot convert instance of {TemporaryFile.__name__} to float")

    def __int__(self) -> None:
        raise ValueError(f"Cannot convert instance of {TemporaryFile.__name__} to int")

    def __len__(self) -> int:
        if not self.length:
            read_bytes = self.read()
            self.seek(0)
            self.length = len(read_bytes)
        return self.length

    def __bool__(self) -> bool:
        return len(self) > 0

    def __str__(self) -> str:

        if not self._str:
            self._str = self.read().decode()

        return self._str

    def __enter__(self) -> IO[Any]:
        return self.file

    def __exit__(self, *args: Any) -> None:
        self.close()


class ParserState(Enum):
    PART_BOUNDARY = 0
    CONTENT_DISPOSITION = 1
    CONTENT_TYPE = 2
    CONTENT_HEADER = 3
    CONTENT_DATA = 4
    END = 5


def parse_multipart_message(data: bytes, boundary: str, encoding: str = "utf8") -> Dict[str, Any]:
    state = ParserState.PART_BOUNDARY
    prev_byte = None
    cursor = 0
    boundary_length = len(boundary)
    string_buffer = ""
    body = {}

    def _append_content_to_body(raw_content_disposition: str, _content_type: str, _content_data: bytes) -> None:
        parsed_content_disposition: Tuple[str, Dict[str, str]] = parse_header(raw_content_disposition[20:])
        if "filename" in parsed_content_disposition[1]:
            tmp_file = TemporaryFile()
            tmp_file.write(_content_data)
            tmp_file.seek(0)
            body[parsed_content_disposition[1]["name"]] = UploadedFile(
                tmp_file,
                _content_type[14:].lower(),
                parsed_content_disposition[1]["filename"],
            )
        else:
            # mypy bug?
            body[parsed_content_disposition[1]["name"]] = _content_data.decode(encoding)  # type: ignore

    content_disposition = ""
    content_type = ""
    content_cursor = 0

    for code in data:
        line_break = code == 0x0A and prev_byte == 0x0D
        new_line_char = code == 0x0A or code == 0x0D

        if not new_line_char:
            string_buffer = string_buffer + chr(code)

        if state is ParserState.PART_BOUNDARY and line_break:
            if string_buffer == "--" + boundary:
                string_buffer = ""
                state = ParserState.CONTENT_DISPOSITION
            else:
                raise IOError("Could not parse message body, body is malformed or incorrect boundary was passed.")

        elif state is ParserState.CONTENT_DISPOSITION and line_break:
            content_disposition = string_buffer
            string_buffer = ""
            state = ParserState.CONTENT_TYPE
            if data[cursor + 1 : cursor + 13].decode(encoding).lower() != "content-type":
                state = ParserState.CONTENT_HEADER

        elif state is ParserState.CONTENT_TYPE and line_break:
            content_type = string_buffer
            state = ParserState.CONTENT_HEADER
            string_buffer = ""

        elif state is ParserState.CONTENT_HEADER and line_break:
            string_buffer = ""
            state = ParserState.CONTENT_DATA
            content_cursor = cursor

        if state is ParserState.CONTENT_DATA:
            if line_break and string_buffer == "--" + boundary:
                content_data = data[content_cursor + 1 : cursor - (boundary_length + 5)]
                _append_content_to_body(content_disposition, content_type, content_data)
                content_type = ""
                state = ParserState.CONTENT_DISPOSITION

            if line_break and string_buffer == "--" + boundary + "--":
                content_data = data[content_cursor + 1 : cursor - (boundary_length + 7)]
                _append_content_to_body(content_disposition, content_type, content_data)
                content_type = ""
                state = ParserState.END

            if line_break:
                string_buffer = ""

        prev_byte = code
        cursor += 1

    return body


__all__ = ["UploadedFile", "parse_multipart_message"]
